@@ -3,6 +3,11 @@ package Net::ACME2;
 use strict;
 use warnings;
 
+our $VERSION;
+BEGIN {
+    $VERSION = '0.40_02';
+}
+
 =encoding utf-8
 
 =head1 NAME
@@ -104,6 +109,8 @@ a new version of this module.
 
 =item * Comprehensive error handling with typed, L<X::Tiny>-based exceptions.
 
+=item * Supports blocking and (experimentally) non-blocking I/O.
+
 =item * L<Retry POST on C<badNonce> errors.|https://tools.ietf.org/html/rfc8555#section-6.5>
 
 =item * This is a pure-Perl solution. Most of its dependencies are
@@ -147,6 +154,36 @@ for more details.
 
 =cut
 
+=head1 EXPERIMENTAL: NON-BLOCKING (ASYNCHRONOUS) I/O
+
+By default, Net::ACME2 uses blocking I/O.
+
+To facilitate asynchronous/non-blocking I/O, you may give an C<async_ua>
+to C<new()>. This value must be an object that implements C<request()>.
+That method should mimic L<HTTP::Tiny>’s method of the same name
+B<except> that, instead of returning a hash reference, it should return
+a promise. (à la L<Promise::XS>, L<Promise::ES6>, L<Mojo::Promise>, etc.)
+That promise’s resolution should be a single value that mimics
+C<HTTP::Tiny::request()>’s return structure.
+
+When a Net::ACME2 instance is created with C<async_ua>, several of the
+methods described below return promises. These promises resolve to the values
+that otherwise would be returned directly in synchronous mode. Any exception
+that would be thrown in synchronous mode is given as the promise’s rejection
+value. This document’s convention to indicate a function that, in
+asynchronous mode, returns a promise is:
+
+    promise($whatever) = ...
+
+This distribution ships with L<Net::ACME2::Curl>, a wrapper around
+L<Net::Curl::Promiser>, which in turns wraps L<Net::Curl::Multi>. This
+provides out-of-the-box support for Perl’s most widely-used event interfaces;
+see Net::Curl::Promiser’s documentation for more details.
+
+=cut
+
+#----------------------------------------------------------------------
+
 use Crypt::Format;
 use MIME::Base64 ();
 
@@ -155,8 +192,7 @@ use Net::ACME2::AccountKey;
 use Net::ACME2::HTTP;
 use Net::ACME2::Order;
 use Net::ACME2::Authorization;
-
-our $VERSION = '0.35_01';
+use Net::ACME2::PromiseUtil;
 
 use constant {
     _HTTP_OK => 200,
@@ -176,6 +212,8 @@ use constant FULL_JWT_METHODS => qw(
     revokeCert
 );
 
+#----------------------------------------------------------------------
+
 =head1 METHODS
 
 =head2 I<CLASS>->new( %OPTS )
@@ -197,6 +235,9 @@ directory contents. Saves a round-trip to the ACME2 server, but there’s
 no built-in logic to determine when the cache goes invalid. Caveat
 emptor.
 
+=item * C<async_ua> - Optional. Provides a custom UA object to facilitate
+non-blocking I/O. This object B<MUST> implement the interface described above.
+
 =back
 
 =cut
@@ -217,6 +258,7 @@ sub _new_without_key_check {
         _key  => $opts{'key'},
         _key_id => $opts{'key_id'},
         _directory => $opts{'directory'},
+        _async_ua => $opts{'async_ua'},
     };
 
     bless $self, $class;
@@ -248,17 +290,21 @@ sub key_id {
 A passthrough interface to the underlying L<HTTP::Tiny> object’s
 C<timeout()> method.
 
+Throws an exception if C<async_ua> was given to C<new()>.
+
 =cut
 
 sub http_timeout {
     my $self = shift;
 
+    die 'Don’t call in asynchronous mode!' if $self->{'_async_ua'};
+
     return $self->{'_http'}->timeout(@_);
 }
 
 #----------------------------------------------------------------------
 
-=head2 $url = I<CLASS>->get_terms_of_service()
+=head2 promise($url) = I<CLASS>->get_terms_of_service()
 
 Returns the URL for the terms of service. Callable as either
 a class method or an instance method.
@@ -273,19 +319,24 @@ sub get_terms_of_service {
         $self = $self->_new_without_key_check();
     }
 
-    my $dir = $self->_get_directory();
+    return Net::ACME2::PromiseUtil::then(
+        $self->_get_directory(),
+        sub {
+            my $dir = shift;
 
-    # Exceptions here indicate an ACME violation and should be
-    # practically nonexistent.
-    my $url = $dir->{'meta'} or _die_generic('No “meta” in directory!');
-    $url = $url->{'termsOfService'} or _die_generic('No “termsOfService” in directory metadata!');
+            # Exceptions here indicate an ACME violation and should be
+            # practically nonexistent.
+            my $url = $dir->{'meta'} or _die_generic('No “meta” in directory!');
+            $url = $url->{'termsOfService'} or _die_generic('No “termsOfService” in directory metadata!');
 
-    return $url;
+            return $url;
+        },
+    );
 }
 
 #----------------------------------------------------------------------
 
-=head2 $created_yn = I<OBJ>->create_account( %OPTS )
+=head2 promise($created_yn) = I<OBJ>->create_account( %OPTS )
 
 Creates an account using the ACME2 object’s key and the passed
 %OPTS, which are as described in the ACME2 spec (cf. C<newAccount>).
@@ -306,34 +357,36 @@ sub create_account {
         ($opts{$name} &&= JSON::true()) ||= JSON::false();
     }
 
-    my $resp = $self->_post(
-        'newAccount',
-        \%opts,
+    return Net::ACME2::PromiseUtil::then(
+        $self->_post( 'newAccount', \%opts ),
+        sub {
+            my ($resp) = @_;
+
+            $self->{'_key_id'} = $resp->header('location');
+
+            $self->{'_http'}->set_key_id( $self->{'_key_id'} );
+
+            return 0 if $resp->status() == _HTTP_OK;
+
+            $resp->die_because_unexpected() if $resp->status() != _HTTP_CREATED;
+
+            my $struct = $resp->content_struct();
+
+            if ($struct) {
+                for my $name (newAccount_booleans()) {
+                    next if !exists $struct->{$name};
+                    ($struct->{$name} &&= 1) ||= 0;
+                }
+            }
+
+            return 1;
+        },
     );
-
-    $self->{'_key_id'} = $resp->header('location');
-
-    $self->{'_http'}->set_key_id( $self->{'_key_id'} );
-
-    return 0 if $resp->status() == _HTTP_OK;
-
-    $resp->die_because_unexpected() if $resp->status() != _HTTP_CREATED;
-
-    my $struct = $resp->content_struct();
-
-    if ($struct) {
-        for my $name (newAccount_booleans()) {
-            next if !exists $struct->{$name};
-            ($struct->{$name} &&= 1) ||= 0;
-        }
-    }
-
-    return 1;
 }
 
 #----------------------------------------------------------------------
 
-=head2 $order = I<OBJ>->create_order( %OPTS )
+=head2 promise($order) = I<OBJ>->create_order( %OPTS )
 
 Returns a L<Net::ACME2::Order> object. %OPTS is as described in the
 ACME spec (cf. C<newOrder>). Boolean values may be given as simple
@@ -348,19 +401,24 @@ sub create_order {
 
     $self->_require_key_id(\%opts);
 
-    my $resp = $self->_post( 'newOrder', \%opts );
+    return Net::ACME2::PromiseUtil::then(
+        $self->_post( 'newOrder', \%opts ),
+        sub {
+            my ($resp) = @_;
 
-    $resp->die_because_unexpected() if $resp->status() != _HTTP_CREATED;
+            $resp->die_because_unexpected() if $resp->status() != _HTTP_CREATED;
 
-    return Net::ACME2::Order->new(
-        id => $resp->header('location'),
-        %{ $resp->content_struct() },
+            return Net::ACME2::Order->new(
+                id => $resp->header('location'),
+                %{ $resp->content_struct() },
+            );
+        },
     );
 }
 
 #----------------------------------------------------------------------
 
-=head2 $authz = I<OBJ>->get_authorization( $URL )
+=head2 promise($authz) = I<OBJ>->get_authorization( $URL )
 
 Fetches the authorization’s information based on the given $URL
 and returns a L<Net::ACME2::Authorization> object.
@@ -372,11 +430,16 @@ The URL is as given by L<Net::ACME2::Order>’s C<authorizations()> method.
 sub get_authorization {
     my ($self, $id) = @_;
 
-    my $resp = $self->_post_as_get($id);
+    return Net::ACME2::PromiseUtil::then(
+        $self->_post_as_get($id),
+        sub {
+            my $resp = shift;
 
-    return Net::ACME2::Authorization->new(
-        id => $id,
-        %{ $resp->content_struct() },
+            return Net::ACME2::Authorization->new(
+                id => $id,
+                %{ $resp->content_struct() },
+            );
+        },
     );
 }
 
@@ -406,7 +469,7 @@ sub make_key_authorization {
 
 #----------------------------------------------------------------------
 
-=head2 I<OBJ>->accept_challenge( $CHALLENGE )
+=head2 promise() = I<OBJ>->accept_challenge( $CHALLENGE )
 
 Signal to the ACME server that the CHALLENGE is ready.
 
@@ -415,19 +478,20 @@ Signal to the ACME server that the CHALLENGE is ready.
 sub accept_challenge {
     my ($self, $challenge_obj) = @_;
 
-    $self->_post_url(
-        $challenge_obj->url(),
-        {
-            keyAuthorization => $self->make_key_authorization($challenge_obj),
-        },
+    return Net::ACME2::PromiseUtil::then(
+        $self->_post_url(
+            $challenge_obj->url(),
+            {
+                keyAuthorization => $self->make_key_authorization($challenge_obj),
+            },
+        ),
+        sub { undef },
     );
-
-    return;
 }
 
 #----------------------------------------------------------------------
 
-=head2 $status = I<OBJ>->poll_authorization( $AUTHORIZATION )
+=head2 promise($status) = I<OBJ>->poll_authorization( $AUTHORIZATION )
 
 Accepts a L<Net::ACME2::Authorization> instance and polls the
 ACME server for that authorization’s status. The $AUTHORIZATION
@@ -442,7 +506,7 @@ As a courtesy, this returns the $AUTHORIZATION’s new C<status()>.
 
 #----------------------------------------------------------------------
 
-=head2 $status = I<OBJ>->finalize_order( $ORDER, $CSR )
+=head2 promise($status) = I<OBJ>->finalize_order( $ORDER, $CSR )
 
 Finalizes an order and updates the $ORDER object with the returned
 status. $CSR may be in either DER or PEM format.
@@ -466,23 +530,28 @@ sub finalize_order {
 
     $csr = MIME::Base64::encode_base64url($csr_der);
 
-    my $post = $self->_post_url(
-        $order_obj->finalize(),
-        {
-            csr => $csr,
+    return Net::ACME2::PromiseUtil::then(
+        $self->_post_url(
+            $order_obj->finalize(),
+            {
+                csr => $csr,
+            },
+        ),
+        sub {
+            my $post = shift;
+
+            my $content = $post->content_struct();
+
+            $order_obj->update($content);
+
+            return $order_obj->status();
         },
     );
-
-    my $content = $post->content_struct();
-
-    $order_obj->update($content);
-
-    return $order_obj->status();
 }
 
 #----------------------------------------------------------------------
 
-=head2 $status = I<OBJ>->poll_order( $ORDER )
+=head2 promise($status) = I<OBJ>->poll_order( $ORDER )
 
 Like C<poll_authorization()> but handles a
 L<Net::ACME2::Order> object instead.
@@ -493,7 +562,7 @@ L<Net::ACME2::Order> object instead.
 
 #----------------------------------------------------------------------
 
-=head2 $cert = I<OBJ>->get_certificate_chain( $ORDER )
+=head2 promise($cert) = I<OBJ>->get_certificate_chain( $ORDER )
 
 Fetches the $ORDER’s certificate chain and returns
 it in the format implied by the
@@ -505,7 +574,12 @@ protocol specification for details about this format.
 sub get_certificate_chain {
     my ($self, $order) = @_;
 
-    return $self->_post_as_get( $order->certificate() )->content();
+    return Net::ACME2::PromiseUtil::then(
+        $self->_post_as_get( $order->certificate() ),
+        sub {
+            return shift()->content();
+        },
+    );
 }
 
 #----------------------------------------------------------------------
@@ -519,18 +593,26 @@ sub _key_thumbprint {
 sub _get_directory {
     my ($self) = @_;
 
-    $self->{'_directory'} ||= do {
+    return $self->{'_directory_promise'} ||= do {
         my $dir_path = $self->DIRECTORY_PATH();
-        $self->{'_http'}->get("https://$self->{'_host'}$dir_path")->content_struct();
+
+        my $http = $self->{'_http'};
+
+        Net::ACME2::PromiseUtil::then(
+            $self->{'_http'}->get("https://$self->{'_host'}$dir_path"),
+            sub {
+                my $dir_hr = shift()->content_struct();
+
+                my $new_nonce_url = $dir_hr->{'newNonce'} or do {
+                    _die_generic('Directory lacks “newNonce”.');
+                };
+
+                $http->set_new_nonce_url( $new_nonce_url );
+
+                return $dir_hr;
+            },
+        );
     };
-
-    my $new_nonce_url = $self->{'_directory'}{'newNonce'} or do {
-        _die_generic('Directory is missing “newNonce”.');
-    };
-
-    $self->{'_http'}->set_new_nonce_url( $new_nonce_url );
-
-    return $self->{'_directory'};
 }
 
 sub _require_key_id {
@@ -546,13 +628,18 @@ sub _require_key_id {
 sub _poll_order_or_authz {
     my ($self, $order_or_authz_obj) = @_;
 
-    my $get = $self->_post_as_get( $order_or_authz_obj->id() );
+    return Net::ACME2::PromiseUtil::then(
+        $self->_post_as_get( $order_or_authz_obj->id() ),
+        sub {
+            my $get = shift;
 
-    my $content = $get->content_struct();
+            my $content = $get->content_struct();
 
-    $order_or_authz_obj->update($content);
+            $order_or_authz_obj->update($content);
 
-    return $order_or_authz_obj->status();
+            return $order_or_authz_obj->status();
+        },
+    );
 }
 
 sub _key_obj {
@@ -567,6 +654,7 @@ sub _set_http {
     $self->{'_http'} = Net::ACME2::HTTP->new(
         key => $self->{'_key'} && $self->_key_obj(),
         key_id => $self->{'_key_id'},
+        ua => $self->{'_async_ua'},
     );
 
     return;
@@ -580,11 +668,18 @@ sub _post {
     my $post_method;
     $post_method = 'post_full_jwt' if grep { $link_name eq $_ } FULL_JWT_METHODS();
 
-    # Since the $link_name will come from elsewhere in this module
-    # there really shouldn’t be an error here, but just in case.
-    my $url = $self->_get_directory()->{$link_name} or _die_generic("Unknown link name: “$link_name”");
+    return Net::ACME2::PromiseUtil::then(
+        $self->_get_directory(),
+        sub {
+            my $dir_hr = shift;
 
-    return $self->_post_url( $url, $data, $post_method );
+            # Since the $link_name will come from elsewhere in this module
+            # there really shouldn’t be an error here, but just in case.
+            my $url = $dir_hr->{$link_name} or _die_generic("Unknown link name: “$link_name”");
+
+            return $self->_post_url( $url, $data, $post_method );
+        },
+    );
 }
 
 sub _post_as_get {
@@ -596,13 +691,18 @@ sub _post_as_get {
 sub _post_url {
     my ( $self, $url, $data, $opt_post_method ) = @_;
 
-    #Do this in case we haven’t initialized the directory yet.
-    #Initializing the directory is necessary to get a nonce.
-    $self->_get_directory();
-
     my $post_method = $opt_post_method || 'post_key_id';
 
-    return $self->{'_http'}->$post_method( $url, $data );
+    my $http = $self->{'_http'};
+
+    #Do this in case we haven’t initialized the directory yet.
+    #Initializing the directory is necessary to get a nonce.
+    return Net::ACME2::PromiseUtil::then(
+        $self->_get_directory(),
+        sub {
+            return $http->$post_method( $url, $data );
+        },
+    );
 }
 
 sub _die_generic {
@@ -613,6 +713,10 @@ sub _die_generic {
 *create_new_account = *create_account;
 *create_new_order = *create_order;
 
+# sub DESTROY {
+#     print "ACME2 destroyed at ${^GLOBAL_PHASE}\n";
+# }
+
 1;
 
 =head1 TODO
